//-*-c++-*-
#ifndef INCLUDED_StateNode_h_
#define INCLUDED_StateNode_h_

#include "Transition.h"
#include "Behaviors/BehaviorBase.h"
#include <vector>
#include <string>

//! Recursive data structure - both a state machine controller as well as a node within a state machine itself
/*! Override setup() to setup your own Transition and StateNode network.*/
class StateNode  : public BehaviorBase {
public:
	//!constructor
	StateNode() : BehaviorBase(), parent(NULL), transitions(), issetup(false), retain(true), nodes(), name("?") {	std::cout << "!" << std::endl;}

	//!constructor, pass a name to use, calls setName(n) 
	StateNode(const char* n, StateNode* p=NULL) : BehaviorBase(), parent(p), transitions(), issetup(false), retain(true), nodes(), name("") {
		setName(n);
	}

	//!destructor, removes references to its outgoing transitions (be careful of incoming ones - they're still around!), and calls RemoveReference() on subnodes
	virtual ~StateNode();

	//!Adds the specified StateTransition to the transition table
	virtual void addTransition(Transition* trans);

	//!Returns the std::vector of transitions so you can modify them yourself if need be
	std::vector<Transition*>& getTransitions() { return transitions; }

	//!Adds a StateNode to #nodes so it can be automatically dereferenced later, returns what it's passed (for convenience), calls AddReference() on @a node
	virtual StateNode* addNode(StateNode* node) { nodes.push_back(node); node->AddReference(); return node; }

	//!Returns the std::vector of nodes so you can modify them yourself if need be
	std::vector<StateNode*>& getNodes() { return nodes; }

	//!Sets the retain flag - if not retained, will RemoveReference() subnodes upon DoStop() and recreate them on DoStart (by calling setup()) - may be too expensive to be worth saving memory...
	void setRetain(bool f) { retain=f; }

	//!Transitions should call this when you are entering the state, so it can enable its transitions
	virtual void DoStart();

	//!This is called by DoStart() when you should setup the network of subnodes
	virtual void setup() {issetup=true;}

	//!Transitions should call this when you are leaving the state, so it can disable its transitions
	virtual void DoStop();
	
	//!This is called by DoStop() when you should destruct subnodes
	virtual void teardown() { std::cout << "Teardown!!!!!!!!" << std::endl;issetup=false;}

	//!set name to @a n.  Makes a copy of @a n, so you can throw it away later.
	void setName(const std::string& n);

	//!returns name of StateNode
	virtual std::string getName() const { return name; }

	//!returns name again (you should override this for top-level nodes to give users better descriptions)
	virtual std::string getDescription() const { return name; }

protected:
	//!called by a subnode when it is being DoStart()'ed
	virtual void transitionTo(StateNode* n);

	//!called by a subnode when it is being DoStop()'ed
	virtual void transitionFrom(StateNode* n);

	//Node Stuff:
	//! pointer to the machine that contains this node
	StateNode* parent;
	//! a vector of outgoing transitions
	std::vector<Transition*> transitions;
	
	//Machine Stuff:
	//! this is set to true if the network has been setup but not destroyed (i.e. don't need to call setupSubNodes again)
	bool issetup;
	//! this is set to true if the network should be retained between activations.  Otherwise it's dereferenced upon DoStop(). (or at least RemoveReference() is called on subnodes)
	bool retain;
	//! vector of StateNodes, just so they can be dereferenced again on DoStop() (unless retained) or ~StateNode()
	std::vector<StateNode*> nodes;

	//Behavior Stuff:
	//! holds the name of the Node/Machine
	std::string name;

private:
	StateNode(const StateNode& node); //!< don't call this
	StateNode operator=(const StateNode& node); //!< don't call this
};

/*! @file
 * @brief Describes StateNode, which is both a state machine controller as well as a node within a state machine itself
 * @author ejt (Creator)
 *
 * $Author: ejt $
 * $Name: tekkotsu-1_5 $
 * $Revision: 1.7 $
 * $State: Rel $
 * $Date: 2003/09/25 15:23:56 $
 */

#endif
